/*
 * byteordr.h -- init / validate byteorder functions.
 * $Id: byteordr.h,v 1.1 2008-01-12 09:46:18 sezero Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef __BYTEORDR_H
#define __BYTEORDR_H

void ValidateByteorder (void);
	/* call this from your main() */

#endif	/* __BYTEORDR_H */

