/* MIDI streaming music support using Timidity. */

#if !defined(_SND_TIMIDITY_H_)
#define _SND_TIMIDITY_H_

#if defined(USE_CODEC_TIMIDITY)

extern snd_codec_t timidity_codec;

#endif	/* USE_CODEC_TIMIDITY */

#endif	/* ! _SND_TIMIDITY_H_ */

