/* MIDI streaming music support using WildMIDI. */

#if !defined(_SND_WILDMIDI_H_)
#define _SND_WILDMIDI_H_

#if defined(USE_CODEC_WILDMIDI)

extern snd_codec_t wildmidi_codec;

#endif	/* USE_CODEC_WILDMIDI */

#endif	/* ! _SND_WILDMIDI_H_ */

