/* DOS sound support for Sound Blaster. */

#if !defined(__HX2_SND_BLASTER)
#define __HX2_SND_BLASTER

#if HAVE_DOS_SB_SOUND

extern snd_driver_t snddrv_blaster;

#endif	/* HAVE_DOS_SB_SOUND */

#endif	/* __HX2_SND_BLASTER */

