
//developer 0

// load the base configuration
exec default.cfg

// load the last saved configuration
exec config.cfg

// run a user script file if present
exec autoexec.cfg

// display the main menu
menu_main

//
// stuff command line statements
//
stuffcmds

// start demos if not already running a server
//startdemos demo1 demo2 demo3

