/* DOS PCI sound card support */

#if !defined(__HX2_SND_PCI)
#define __HX2_SND_PCI

#if HAVE_DOS_PCI_SOUND

extern snd_driver_t snddrv_pci;

#endif	/* HAVE_DOS_PCI_SOUND */

#endif	/* __HX2_SND_PCI */

