/* DOS sound support for GUS. */

#if !defined(__HX2_SND_GUS)
#define __HX2_SND_GUS

#if HAVE_DOS_GUS_SOUND

extern snd_driver_t snddrv_gus;

#endif	/* HAVE_DOS_GUS_SOUND */

#endif	/* __HX2_SND_GUS */

