/* alternative SDL-using implementations of system functions needed by
 * common sys_unix.c, when platform specific ones are not available. */

#ifndef SYS_SDL_H
#define SYS_SDL_H

/* NOTHING HERE YET */

#endif /* SYS_SDL_H */

