// win32res.rc -- Win32 resource file for Hexen II
// $Id: win32res.rc 4767 2012-06-16 20:48:51Z sezero $

#include <afxres.h>
#include <winresrc.h>
#include "resource.h"

// Icon
IDI_ICON2 ICON DISCARDABLE RES_ICONFILE


// Dialog
#if !defined(NO_SPLASHES)

#if !defined(GLQUAKE)

IDD_DIALOG1 DIALOGEX 0, 0, 78, 21
STYLE DS_MODALFRAME | DS_SETFOREGROUND | DS_3DLOOK | DS_CENTER | WS_POPUP | WS_VISIBLE
EXSTYLE WS_EX_TOOLWINDOW | WS_EX_CLIENTEDGE
FONT 16, "Times New Roman", 0, 0, 0x1
BEGIN
    CTEXT SPLASH_STR, IDC_STATIC, 2, 3, 74, 15, SS_CENTERIMAGE | SS_SUNKEN
END

#else	// OpenGL version

IDD_DIALOG1 DIALOGEX 0, 0, 167, 27
STYLE DS_MODALFRAME | DS_SETFOREGROUND | DS_3DLOOK | DS_CENTER | WS_POPUP | WS_VISIBLE
EXSTYLE WS_EX_TOOLWINDOW | WS_EX_CLIENTEDGE
FONT 16, "Times New Roman", 0, 0, 0x1
BEGIN
    CONTROL 112, IDC_STATIC, "Static", SS_BITMAP | SS_REALSIZEIMAGE, 0, 0, 167, 28
END

// Bitmap
IDB_HXBITMAP BITMAP DISCARDABLE SPLASH_BMP

#endif	// GLQUAKE

#endif	// NO_SPLASHES


// String Table
STRINGTABLE DISCARDABLE
BEGIN
	IDS_STRING1 RES_STRING
END

