/* Sound support using SunAudio. */

#if !defined(__HX2_SND_SUN)
#define __HX2_SND_SUN

#if HAVE_SUN_SOUND

extern snd_driver_t snddrv_sunaudio;

#endif	/* HAVE_SUN_SOUND */

#endif	/* __HX2_SND_SUN */

