/* Sound support using Windows multimedia api */

#if !defined(__HX2_SND_WIN)
#define __HX2_SND_WIN

#if HAVE_WIN_SOUND

extern snd_driver_t snddrv_win;

#endif	/* HAVE_WIN_SOUND */

#endif	/* __HX2_SND_WIN */

