/* Sound support using Windows DirectSound. */

#if !defined(__HX2_SND_DSOUND)
#define __HX2_SND_DSOUND

#if HAVE_WIN_DX_SOUND

extern snd_driver_t snddrv_dsound;

#endif	/* HAVE_WIN_DX_SOUND */

#endif	/* __HX2_SND_DSOUND */

