/* Sound support using ALSA. */

#if !defined(__HX2_SND_ALSA)
#define __HX2_SND_ALSA

#if HAVE_ALSA_SOUND

extern snd_driver_t snddrv_alsa;

#endif	/* HAVE_ALSA_SOUND */

#endif	/* __HX2_SND_ALSA */

